/*    [-*- C++ -*-] */
/*
THE COMPUTER CODE CONTAINED HEREIN IS THE SOLE PROPERTY OF PARALLAX
SOFTWARE CORPORATION ("PARALLAX").  PARALLAX, IN DISTRIBUTING THE CODE TO
END-USERS, AND SUBJECT TO ALL OF THE TERMS AND CONDITIONS HEREIN, GRANTS A
ROYALTY-FREE, PERPETUAL LICENSE TO SUCH END-USERS FOR USE BY SUCH END-USERS
IN USING, DISPLAYING,  AND CREATING DERIVATIVE WORKS THEREOF, SO LONG AS
SUCH USE, DISPLAY OR CREATION IS FOR NON-COMMERCIAL, ROYALTY OR REVENUE
FREE PURPOSES.  IN NO EVENT SHALL THE END-USER USE THE COMPUTER CODE
CONTAINED HEREIN FOR REVENUE-BEARING PURPOSES.  THE END-USER UNDERSTANDS
AND AGREES TO THE TERMS HEREIN AND ACCEPTS THE SAME BY USE OF THIS FILE.  
COPYRIGHT 1993-1998 PARALLAX SOFTWARE CORPORATION.  ALL RIGHTS RESERVED.
*/
#include "gr.h"
#include "grdef.h"
#include "error.h"
#include "byteswap.h"

#include "rle.h"

int gr_bitblt_dest_step_shift = 0;
int gr_bitblt_double = 0;
ubyte *gr_bitblt_fade_table=NULL;

#include "debug.h"
#include <stdio.h> /* Only for debugging output */

//#define COPY_WITH_FP_UNIT /* Slower for my ppro with pixel doubling, faster on 640x480 sometimes (code align?), namely 18.6 instead of 17.5 fps at level 1 start (full detail) */
//#define PIXDOUB_COPY_WITH_FP_UNIT


#define THRESHOLD	8

#if !(defined(USE_2D_ASM) )
void gr_linear_movsd(ubyte * src, ubyte * dest, uint num_pixels )	
{
#ifdef COPY_WITH_FP_UNIT
	int i;
	uint n, r;
	double *d, *s;
	ubyte *d1, *s1;

// check to see if we are starting on an even byte boundry
// if not, move appropriate number of bytes to even
// 8 byte boundry

	if (num_pixels < THRESHOLD) {
		for (i = 0; i < num_pixels; i++)
			*dest++ = *src++;
		return;
	}

	i = 0;
	if ((r = (int)src & 0x7)) {
		for (i = 0; i < 8 - r; i++)
			*dest++ = *src++;
	}
	num_pixels -= i;

//	n = num_pixels / 8;
//	r = num_pixels % 8;
//WHS: my replacement:
	n = num_pixels >>3;
	r = num_pixels & 7;
#if 0
	// ALEX DID THIS
	r = num_pixels - (n * 8);
#endif
	s = (double *)src;
	d = (double *)dest;
	for (i = 0; i < n; i++)
		*d++ = *s++;
	s1 = (ubyte *)s;
	d1 = (ubyte *)d;
	for (i = 0; i < r; i++)
		*d1++ = *s1++;
#else /* Don't use FP to store 64 bit */
//	int i;
//	for (i = num_pixels; i; i--)
//		*dest++ = *src++;

	memcpy(dest, src, num_pixels); //looks slightly faster (but sometimes slower than copying with FP on the ppro)

#endif
}
#endif	//!( defined(USE_2D_ASM) )

void gr_linear_movsd_double(ubyte *src, ubyte *dest, int num_pixels)
{
#ifdef PIXDOUB_COPY_WITH_FP_UNIT
#define COPY_V5
#ifdef COPY_V5
	double	*d = (double *)dest;
	ubyte	*s = (ubyte *)src;
//	uint	doubletemp[2];
ubyte *doubletemp;
//	uint	temp1, temp2, work;
	uint   work;
	uint   temp;
	int		i;
double ttt;

doubletemp=&ttt;
	for (i = 0; i < num_pixels>>2; i++) {
		work = *s++;
		doubletemp[0]=work; doubletemp[1]=work;
		work = *s++;
		doubletemp[2]=work; doubletemp[3]=work;
		work = *s++;
		doubletemp[4]=work; doubletemp[5]=work;
		work = *s++;
		doubletemp[6]=work; doubletemp[7]=work;

		*d = *(double *) &(doubletemp[0]);
		d++;
	}
#endif
#ifdef COPY_V4
	double	*d = (double *)dest;
	ubyte	*s = (ubyte *)src;
//	uint	doubletemp[2];
uint *doubletemp;
//	uint	temp1, temp2, work;
	uint   work;
	uint   temp;
	int		i;
double ttt;

doubletemp=&ttt;
	for (i = 0; i < num_pixels>>2; i++) {
		work = *s++;
temp=work;
temp |=work<<8;
		work = *s++;
temp |=work<<16;
temp |=work<<24;

#ifdef LITTLE_ENDIAN_MACHINE
		doubletemp[0] = temp;
#else
		doubletemp[0] = temp;
#endif
		work = *s++;
temp=work;
temp |=work<<8;
		work = *s++;
temp |=work<<16;
temp |=work<<24;

#ifdef LITTLE_ENDIAN_MACHINE
		doubletemp[1] = temp;
#else
		doubletemp[1] = temp;
#endif

		*d = *(double *) &(doubletemp[0]);
		d++;
	}
#endif
#ifdef COPY_V3 /* sp 42 */ faster than V1 on the ppro */
	double	*d = (double *)dest;
	uint	*s = (uint *)src;
	uint	*doubletemp;
char a[16];

	uint	temp1, temp2, work;
	int		i;
doubletemp=((long) &a[0]) -( ((long) &a[0]) &7 ) +8;

	for (i = 0; i < num_pixels>>2; i++) {
		work = *s++;

		temp1 = (work &0xFF000000);
		temp1 |= temp1>>8;
		temp2 = (work &0x00FF0000)>>8;
		temp2 |= (temp2>>8);
#ifdef LITTLE_ENDIAN_MACHINE
//		doubletemp[1] = temp2 | (temp2>>8) | temp1 | (temp1>>8);
		doubletemp[1] = temp2 | temp1;
#else
		doubletemp[0] = temp2 | temp1;
#endif
		temp1 = (work &0x0000FF00)<<8;
		temp1 |= temp1<<8;
		temp2 = (work &0x000000FF);
		temp2 |= (temp2<<8);
#ifdef LITTLE_ENDIAN_MACHINE
//		doubletemp[0] = temp2 | (temp2<<8) | temp1 | (temp1<<8);
		doubletemp[0] = temp2 | temp1;

#else
		doubletemp[1] = temp2 | temp1;
#endif

		*d = *(double *) &(doubletemp[0]);
		d++;
	}
#endif
#ifdef COPY_V2 /* sp 28 */
	double	*d = (double *)dest;
	ushort	*s = (ushort *)src;
	ushort	doubletemp[4];
	ushort	temp, work;
	int		i;

	for (i = 0; i < num_pixels >> 2; i++) {
		temp = work = *s++;

		temp=(temp &0xff00) | ((temp>>8) &0xff);
doubletemp[1]=temp;
		temp=(work <<8) | (work &0xff);
doubletemp[0]=temp;
		temp = work = *s++;
		temp=(temp &0xff00) | ((temp>>8) &0xff);
doubletemp[3]=temp;
		temp=(work <<8) | (work &0xff);
doubletemp[2]=temp;

		*d = *(double *) &(doubletemp[0]);
		d++;
	}
#endif
#ifdef COPY_V1 /* sp 41 */
	double	*d = (double *)dest;
	uint	*s = (uint *)src;
//	uint	doubletemp[2];
	uint	*doubletemp;
double ttt;
	uint	temp, work;
	int		i;
doubletemp=&ttt;
	for (i = 0; i < num_pixels>>2; i++) {
		temp = work = *s++;

		temp = ((temp >> 8) & 0x00FFFF00) | (temp & 0xFF0000FF); // 0xABCDEFGH -> 0xABABCDEF
		temp = ((temp >> 8) & 0x000000FF) | (temp & 0xFFFFFF00); // 0xABABCDEF -> 0xABABCDCD
#ifdef LITTLE_ENDIAN_MACHINE
		doubletemp[1] = temp;
#else
		doubletemp[0] = temp;
#endif

		work = ((work << 8) & 0x00FFFF00) | (work & 0xFF0000FF); // 0xABCDEFGH -> 0xABEFGHGH
		work = ((work << 8) & 0xFF000000) | (work & 0x00FFFFFF); // 0xABEFGHGH -> 0xEFEFGHGH
#ifdef LITTLE_ENDIAN_MACHINE
		doubletemp[0] = work;
#else
		doubletemp[1] = work;
#endif

		*d = *(double *) &(doubletemp[0]);
		d++;
	}
#endif
#else

#define MOVSD_DOUBLE_1

#  ifdef MOVSD_DOUBLE_1
	int i;
	for (i=num_pixels; i; i--) {
		*dest++ = *src;
		*dest++ = *src++;
	}
#  endif
#  ifdef MOVSD_DOUBLE_2 /* Considerably slower on my ppro */
        short int *dst = (short int *) dest;
        short int buf;
        int  i;
        for (i=num_pixels; i; i--) {
                buf= *src++; buf |= (buf<<8);
                *dst++ = buf;
        }
#  endif
#  ifdef MOVSD_DOUBLE_3 /* Better but still slower on my ppro */
        int *dst = (short int *) dest;
		int buf;
		int tmp;
		byte *bbuf=&buf;

        int  i;
        for (i=num_pixels>>1; i; i--) { /* note one pixel may be skipped, but this is just a test .. */
                tmp= *src++; buf =tmp| (tmp<<8);
                tmp= *src++; buf |= (tmp<<16) | (tmp<<24);
                *dst++ = buf;
        }
#  endif

#endif
}

void gr_linear_rep_movsdm(ubyte * src, ubyte * dest, int num_pixels )	
{
	int i;
	for (i=0; i<num_pixels; i++ )	{
		if (*src != TRANSPARENCY_COLOR )
			*dest = *src;
		dest++;
		src++;	
	}
}

void gr_linear_rep_movsdm_faded(ubyte * src, ubyte * dest, uint num_pixels, ubyte fade_value )	
{
	int i;
	ubyte source;
	ubyte *fade_base;

	fade_base = gr_fade_table + (fade_value * 256);

	for (i=num_pixels; i != 0; i-- )
	{
		source = *src;
		if (source != (ubyte)TRANSPARENCY_COLOR )
			*dest = *(fade_base + source);
		dest++;
		src++;	
	}

#if 0
	ubyte c;
	ubyte *new_fade_table = &gr_fade_table[fade_value*256];
	uint i;

	for (i=0; i<num_pixels; i++ )	{
		c = *src;
		if (c != TRANSPARENCY_COLOR )
			*dest = new_fade_table[c];
		dest++;
		src++;	
	}
#endif
}

void gr_bm_ubitblt_rle(int w, int h, int dx, int dy, int sx, int sy, grs_bitmap * src, grs_bitmap * dest)
{
	unsigned char * dbits;
	unsigned char * sbits;

	int i;
DEB1(prt("gr_bm_ubitblt_rle\n"));
	sbits = &src->bm_data[4 + src->bm_h];
	for (i=0; i<sy; i++ )
		sbits += (int)src->bm_data[4+i];

	dbits = dest->bm_data + (dest->bm_rowsize * dy) + dx;

	// No interlacing, copy the whole buffer.
	for (i=0; i < h; i++ )    {
		gr_rle_expand_scanline( dbits, sbits, sx, sx+w-1 );
		sbits += (int)src->bm_data[4+i+sy];
		dbits += dest->bm_rowsize << gr_bitblt_dest_step_shift;
	}
}


//WHS: Dit werkt alleen voor MAC, PC moet pixel doubling etc doen (?)
void gr_bitblt_cockpit(grs_bitmap *bm)
{
	unsigned char *dbits;
	unsigned char *sbits;
	int i, h, w;
	ushort s;
DEB1(prt("2d/bitblt.c: gr_bitblt_cockpit\n"));
	w = bm->bm_w;
	h = bm->bm_h;
DEB1(prt("2d/bitblt.c: gr_bitblt_cockpit: w=%d, h=%d\n", bm->bm_w, bm->bm_h));
	sbits = &bm->bm_data[4 + (2 * h)];			// integer size at beginning of bitmap
	dbits = grd_curcanv->cv_bitmap.bm_data;
	for (i = 0; i < h; i++) {
DEB1(prt("2d/bitblt.c: gr_bitblt_cockpit, h=%d\n", i));
		gr_rle_expand_scanline( dbits, sbits, 0, w-1 );
		s = swapshort(*((short *)&(bm->bm_data[4+(i*2)])));
		sbits += (int)s;
		dbits += (int)grd_curcanv->cv_bitmap.bm_rowsize;
	}
}

void gr_bm_ubitbltm_rle(int w, int h, int dx, int dy, int sx, int sy, grs_bitmap * src, grs_bitmap * dest)
{
	unsigned char * dbits;
	unsigned char * sbits;

	int i;
DEB1(prt("gr_bm_ubitbltm_rle\n"));
	sbits = &src->bm_data[4 + src->bm_h];
	for (i=0; i<sy; i++ )
		sbits += (int)src->bm_data[4+i];

	dbits = dest->bm_data + (dest->bm_rowsize * dy) + dx;

	// No interlacing, copy the whole buffer.
	for (i=0; i < h; i++ )    {
		gr_rle_expand_scanline_masked( dbits, sbits, sx, sx+w-1 );
		sbits += (int)src->bm_data[4+i+sy];
		dbits += dest->bm_rowsize << gr_bitblt_dest_step_shift;
	}
}

void gr_ubitmap( int x, int y, grs_bitmap *bm )
{
	register int y1;
	int dest_rowsize;

	unsigned char * dest;
	unsigned char * src;

	if ( bm->bm_flags & BM_FLAG_RLE )	{
		gr_bm_ubitblt_rle(bm->bm_w, bm->bm_h, x, y, 0, 0, bm, &grd_curcanv->cv_bitmap );
		return;
	}

	dest_rowsize=grd_curcanv->cv_bitmap.bm_rowsize << gr_bitblt_dest_step_shift;
	dest = &(grd_curcanv->cv_bitmap.bm_data[ dest_rowsize*y+x ]);

	src = bm->bm_data;

	for (y1=0; y1 < bm->bm_h; y1++ )    {
		gr_linear_movsd( src, dest, bm->bm_w );
		src += bm->bm_rowsize;
		dest+= (int)(dest_rowsize);
	}
}

void gr_ubitmapm( int x, int y, grs_bitmap *bm )
{
	register int y1;
	int dest_rowsize;

	unsigned char * dest;
	unsigned char * src;

	if ( bm->bm_flags & BM_FLAG_RLE )	{
		gr_bm_ubitbltm_rle(bm->bm_w, bm->bm_h, x, y, 0, 0, bm, &grd_curcanv->cv_bitmap );
		return;
	}

	dest_rowsize=grd_curcanv->cv_bitmap.bm_rowsize << gr_bitblt_dest_step_shift;
	dest = &(grd_curcanv->cv_bitmap.bm_data[ dest_rowsize*y+x ]);

	src = bm->bm_data;

	if (gr_bitblt_fade_table==NULL)	{
		for (y1=0; y1 < bm->bm_h; y1++ )    {
			gr_linear_rep_movsdm( src, dest, bm->bm_w );
			src += bm->bm_rowsize;
			dest+= (int)(dest_rowsize);
		}
	} else {
		for (y1=0; y1 < bm->bm_h; y1++ )    {
			gr_linear_rep_movsdm_faded( src, dest, bm->bm_w, gr_bitblt_fade_table[y1+y] );
			src += bm->bm_rowsize;
			dest+= (int)(dest_rowsize);
		}
	}
}

extern void BlitLargeAlign(ubyte *draw_buffer, int dstRowBytes, ubyte *dstPtr, int w, int h, int modulus);

#include "macsys.h" /* For GGI stuff */

void gr_bm_ubitblt_double(int w, int h, int dx, int dy, int sx, int sy, grs_bitmap *src, grs_bitmap *dest)
{
	ubyte * dbits;
	ubyte * sbits;
	int dstep, i;

//printf("gr_bm_ubitblt_double: w,h,=%d,%d,  sx,sy=%d,%d,  dx,dy=%d,%d\n", w,h,sx,sy,dx,dy);

	sbits = src->bm_data  + (src->bm_rowsize * sy) + sx;
	dbits = dest->bm_data + (dest->bm_rowsize * dy) + dx;
	dstep = dest->bm_rowsize << gr_bitblt_dest_step_shift;
#if defined(__powerc) || ( defined(USE_2D_ASM) )
	BlitLargeAlign(sbits, dstep, dbits, src->bm_w, src->bm_h, src->bm_rowsize);
#else

#if 0
//	if (dest->bm_data==plb->write) {
	if (0) {
		for (i=0; i < h; i++ )    {
			gr_linear_movsd_double(sbits, dbits, w);
			dbits += dstep;
//			gr_linear_movsd_double(sbits, dbits, w);
			dbits += dstep;
			sbits += src->bm_rowsize;
		}
		for (i=0; i < h; i++ )    {
			ggiCopyBox(visual, sx, sy+i*2, w*2, 1, dx, dy+i*2+1);
		}
	}
else
#endif
	for (i=0; i < h; i++ )    {
		gr_linear_movsd_double(sbits, dbits, w);
		dbits += dstep;
//WHS:
			gr_linear_movsd_double(sbits, dbits, w);
//memcpy(dbits, dbits-dstep, w*2); /* WHS: this is not faster under KGI! (when screen ram<-> screen ram ) */
		dbits += dstep;
		sbits += src->bm_rowsize;
	}
#endif
}

// From Linear to Linear
void gr_bm_ubitblt(int w, int h, int dx, int dy, int sx, int sy, grs_bitmap * src, grs_bitmap * dest)
{
	unsigned char * dbits;
	unsigned char * sbits;
	//int	src_bm_rowsize_2, dest_bm_rowsize_2;
	int dstep;

	int i;

	if ( src->bm_flags & BM_FLAG_RLE )	{
		gr_bm_ubitblt_rle( w, h, dx, dy, sx, sy, src, dest );
		return ;
	}

	sbits =   src->bm_data  + (src->bm_rowsize * sy) + sx;
	dbits =   dest->bm_data + (dest->bm_rowsize * dy) + dx;

	dstep = dest->bm_rowsize << gr_bitblt_dest_step_shift;

	// No interlacing, copy the whole buffer.
	for (i=0; i < h; i++ )    {
		gr_linear_movsd( sbits, dbits, w );
		sbits += src->bm_rowsize;
		dbits += dstep;
	}
}

// From Linear to Linear Masked
void gr_bm_ubitbltm(int w, int h, int dx, int dy, int sx, int sy, grs_bitmap * src, grs_bitmap * dest)
{
	unsigned char * dbits;
	unsigned char * sbits;

	int i;

	sbits =   src->bm_data  + (src->bm_rowsize * sy) + sx;
	dbits =   dest->bm_data + (dest->bm_rowsize * dy) + dx;

	// No interlacing, copy the whole buffer.

	if (gr_bitblt_fade_table==NULL)	{
		for (i=0; i < h; i++ )    {
			gr_linear_rep_movsdm( sbits, dbits, w );
			sbits += src->bm_rowsize;
			dbits += dest->bm_rowsize;
		}
	} else {
		for (i=0; i < h; i++ )    {
			gr_linear_rep_movsdm_faded( sbits, dbits, w, gr_bitblt_fade_table[dy+i] );
			sbits += src->bm_rowsize;
			dbits += dest->bm_rowsize;
		}
	}
}


void gr_bm_bitblt(int w, int h, int dx, int dy, int sx, int sy, grs_bitmap * src, grs_bitmap * dest)
{
	int dx1=dx, dx2=dx+dest->bm_w-1;
	int dy1=dy, dy2=dy+dest->bm_h-1;

	int sx1=sx, sx2=sx+src->bm_w-1;
	int sy1=sy, sy2=sy+src->bm_h-1;

	if ((dx1 >= dest->bm_w ) || (dx2 < 0)) return;
	if ((dy1 >= dest->bm_h ) || (dy2 < 0)) return;
	if ( dx1 < 0 ) { sx1 += -dx1; dx1 = 0; }
	if ( dy1 < 0 ) { sy1 += -dy1; dy1 = 0; }
	if ( dx2 >= dest->bm_w )	{ dx2 = dest->bm_w-1; }
	if ( dy2 >= dest->bm_h )	{ dy2 = dest->bm_h-1; }

	if ((sx1 >= src->bm_w ) || (sx2 < 0)) return;
	if ((sy1 >= src->bm_h ) || (sy2 < 0)) return;
	if ( sx1 < 0 ) { dx1 += -sx1; sx1 = 0; }
	if ( sy1 < 0 ) { dy1 += -sy1; sy1 = 0; }
	if ( sx2 >= src->bm_w )	{ sx2 = src->bm_w-1; }
	if ( sy2 >= src->bm_h )	{ sy2 = src->bm_h-1; }

	// Draw bitmap bm[x,y] into (dx1,dy1)-(dx2,dy2)
	if ( dx2-dx1+1 < w )
		w = dx2-dx1+1;
	if ( dy2-dy1+1 < h )
		h = dy2-dy1+1;
	if ( sx2-sx1+1 < w )
		w = sx2-sx1+1;
	if ( sy2-sy1+1 < h )
		h = sy2-sy1+1;

	gr_bm_ubitblt(w,h, dx1, dy1, sx1, sy1, src, dest );
}


// Clipped bitmap ... 

void gr_bitmap( int x, int y, grs_bitmap *bm )
{
	int dx1=x, dx2=x+bm->bm_w-1;
	int dy1=y, dy2=y+bm->bm_h-1;
	int sx=0, sy=0;

	if ((dx1 >= grd_curcanv->cv_bitmap.bm_w ) || (dx2 < 0)) return;
	if ((dy1 >= grd_curcanv->cv_bitmap.bm_h) || (dy2 < 0)) return;
	if ( dx1 < 0 ) { sx = -dx1; dx1 = 0; }
	if ( dy1 < 0 ) { sy = -dy1; dy1 = 0; }
	if ( dx2 >= grd_curcanv->cv_bitmap.bm_w )	{ dx2 = grd_curcanv->cv_bitmap.bm_w-1; }
	if ( dy2 >= grd_curcanv->cv_bitmap.bm_h )	{ dy2 = grd_curcanv->cv_bitmap.bm_h-1; }

	// Draw bitmap bm[x,y] into (dx1,dy1)-(dx2,dy2)

	gr_bm_ubitblt(dx2-dx1+1,dy2-dy1+1, dx1, dy1, sx, sy, bm, &grd_curcanv->cv_bitmap );
}

void gr_bitmapm( int x, int y, grs_bitmap *bm )
{
	int dx1=x, dx2=x+bm->bm_w-1;
	int dy1=y, dy2=y+bm->bm_h-1;
	int sx=0, sy=0;

	if ((dx1 >= grd_curcanv->cv_bitmap.bm_w ) || (dx2 < 0)) return;
	if ((dy1 >= grd_curcanv->cv_bitmap.bm_h) || (dy2 < 0)) return;
	if ( dx1 < 0 ) { sx = -dx1; dx1 = 0; }
	if ( dy1 < 0 ) { sy = -dy1; dy1 = 0; }
	if ( dx2 >= grd_curcanv->cv_bitmap.bm_w )	{ dx2 = grd_curcanv->cv_bitmap.bm_w-1; }
	if ( dy2 >= grd_curcanv->cv_bitmap.bm_h )	{ dy2 = grd_curcanv->cv_bitmap.bm_h-1; }

	// Draw bitmap bm[x,y] into (dx1,dy1)-(dx2,dy2)
	if ( bm->bm_flags & BM_FLAG_RLE )	
		gr_bm_ubitbltm_rle(dx2-dx1+1,dy2-dy1+1, dx1, dy1, sx, sy, bm, &grd_curcanv->cv_bitmap );
	else
		gr_bm_ubitbltm(dx2-dx1+1,dy2-dy1+1, dx1, dy1, sx, sy, bm, &grd_curcanv->cv_bitmap );
}
